from typing import Union, Sequence, List, Dict

import dash_core_components as dcc
import dash_html_components as html
import dash
from dash.dependencies import Output, Input
from dash.development.base_component import Component

from pyfileconfgui.component import PFCGuiComponent
from pyfileconfgui.dash_ext.component import DashPythonComponent
from pyfileconfgui.dash_ext.query import get_triggering_id
from pyfileconfgui.dash_ext.tb import TracebackComponent
from pyfileconfgui.pages.navigator.refresh import is_refresh_trigger, REFRESH_INTERVAL_ID
from pyfileconfgui.pfc.reformat import convert_name_to_id


class ReloadPFCComponent(PFCGuiComponent):
    _should_auto_update: bool = True
    _pm_id_name_map: Dict[str, str]

    def __init__(self, id: str):
        self._pm_id_name_map = {}
        super().__init__(id)

    @property
    def layout(self) -> Sequence[Union[DashPythonComponent, Component]]:
        return [
            *self.buttons,
            html.Div(id='reload-output'),
        ]

    @property
    def buttons(self) -> List[html.Button]:
        buttons = []
        for manager in self.gui.runner.managers.values():
            id_name = convert_name_to_id(manager.name)
            id_ = f'reload-button-{id_name}'
            self._pm_id_name_map[id_] = manager.name
            button = html.Button(f'Reload {manager.name}', id=id_)
            buttons.append(button)

        return buttons

    @property
    def button_ids(self) -> List[str]:
        return [comp.id for comp in self.buttons]

    @property
    def button_inputs(self) -> List[Input]:
        return [Input(id_, 'n_clicks') for id_ in self.button_ids]

    def add_callbacks(self, app: dash.Dash) -> None:
        self.add_callback(
            app,
            self.reload_pfc,
            Output('reload-output', 'children'),
            [*self.button_inputs, Input(REFRESH_INTERVAL_ID, 'n_intervals')]
        )
        super().add_callbacks(app)

    def reload_pfc(self, *args):
        if is_refresh_trigger():
            if self._should_auto_update:
                return ''
            return dash.no_update

        trigger_id = get_triggering_id()
        manager_name = self._pm_id_name_map[trigger_id]
        try:
            self.gui.runner.reload([manager_name])
            self._should_auto_update = True
        except Exception as e:
            self._should_auto_update = False
            return TracebackComponent('reload-pfc-traceback').component
        return f'Reloaded pipeline manager {manager_name}'





from typing import Union, Sequence

from dash.development.base_component import Component
import dash_core_components as dcc

from pyfileconfgui.dash_ext.component import DashPythonComponent


class PythonBlockComponent(DashPythonComponent):

    def __init__(self, id: str, content: str, **kwargs):
        self.content = content
        super().__init__(id, **kwargs)

    @property
    def layout(self) -> Sequence[Union['DashPythonComponent', Component]]:
        return [
            dcc.Markdown(f'```python\n{self.content}\n```', style={'overflow': 'auto'})
        ]


import dash

external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']


def create_app() -> dash.Dash:
    app = dash.Dash('pyfileconf', external_stylesheets=external_stylesheets)
    app.config.suppress_callback_exceptions = True
    return app



from typing import Dict, Any, Optional, Union, Callable, Type, Sequence

from pyfileconf import Selector, PipelineManager


class PFCRunner:

    def __init__(self):
        self.s = Selector()

    def run(self, path: str) -> Any:
        path_parts = path.split('/')
        if len(path_parts) < 2:
            raise ValueError(f'got invalid path {path}')
        manager_name = path_parts[0]
        run_path = '.'.join(path_parts[1:])
        manager = self.managers[manager_name]
        return manager.run(run_path)

    def create(self, section_path: str, func_or_class: Optional[Union[Callable, Type]] = None):
        path_parts = section_path.split('.')
        if len(path_parts) < 2:
            raise ValueError(f'got invalid section path {section_path}')
        manager_name = path_parts[0]
        create_path = '.'.join(path_parts[1:])
        manager = self.managers[manager_name]
        return manager.create(create_path, func_or_class)

    @property
    def managers(self) -> Dict[str, PipelineManager]:
        return self.s._managers

    def reload(self, manager_names: Optional[Sequence[str]] = None):
        for manager in self.managers.values():
            if manager_names and manager.name not in manager_names:
                continue
            manager.reload()




from pyfileconf import Selector

from pyfileconfgui import PyFileConfGUI
from tests.pfc import full_pm_setup

if __name__ == '__main__':
    pm = full_pm_setup()
    gui = PyFileConfGUI()
    gui.run_server(debug=True, use_reloader=False)


from typing import TYPE_CHECKING, Dict, Union, Sequence

from dash.dependencies import Output, Input
from dash.development.base_component import Component

from pyfileconfgui.component import PFCGuiComponent
from pyfileconfgui.dash_ext.component import DashPythonComponent
from pyfileconfgui.pages.navigator.create import CreateEntryComponent
from pyfileconfgui.pages.navigator.edit import EditItemComponent
from pyfileconfgui.pages.navigator.refresh import REFRESH_INTERVAL_ID
from pyfileconfgui.pages.navigator.reload import ReloadPFCComponent
from pyfileconfgui.pages.navigator.run import RunEntryComponent


import dash_core_components as dcc
import dash_html_components as html
from dash_keyed_file_browser import KeyedFileBrowser
from dash import dash


def show_running_item(selected_file: Dict[str, str]):
    if not selected_file:
        return dash.no_update
    path = selected_file['key']
    return path


def show_editing_item(selected_file: Dict[str, str]):
    if not selected_file:
        return dash.no_update
    path = selected_file['key']
    return path


class NavigatorComponent(PFCGuiComponent):

    @property
    def layout(self) -> Sequence[Union['DashPythonComponent', Component]]:
        app = self.gui.app

        layout = [
            html.Label('Pyfileconf Items'),
            dcc.Interval(REFRESH_INTERVAL_ID, 2000),
            ReloadPFCComponent('reload-pfc-root'),
            KeyedFileBrowser(self.gui.file_objs, id='kfb'),
            CreateEntryComponent('create-item-root'),
            RunEntryComponent('run-item-root'),
            EditItemComponent('edit-item-root'),
        ]

        return layout

    def add_callbacks(self, app: dash.Dash) -> None:
        self.add_callback(
            app,
            show_running_item,
            Output('run-input', 'children'),
            [Input('kfb', 'openFile')]
        )
        self.add_callback(
            app,
            self.update_files_after_creating_item,
            Output('kfb', 'files'),
            [Input('create-item-output', 'children')]
        )
        self.add_callback(
            app,
            show_editing_item,
            Output('edit-item-name-output', 'children'),
            [Input('kfb', 'selectedFile')]
        )
        super().add_callbacks(app)

    def update_files_after_creating_item(self, updated_message: str):
        self.gui.refresh()
        return self.gui.file_objs






import traceback

from pyfileconfgui.dash_ext.python import PythonBlockComponent

DEFAULT_HEIGHT = '300px'


class TracebackComponent(PythonBlockComponent):

    def __init__(self, id: str, **kwargs):
        if 'style' not in kwargs:
            kwargs['style'] = {'max-height': DEFAULT_HEIGHT}
        elif 'max-height' not in kwargs['style']:
            kwargs['style']['max-height'] = DEFAULT_HEIGHT
        content = traceback.format_exc()
        super().__init__(id, content, **kwargs)

    def refresh(self):
        self.content = traceback.format_exc()



import traceback
from io import StringIO
from typing import Union, Sequence
import sys

import dash
from dash.dependencies import Output, Input
from dash.development.base_component import Component
import dash_core_components as dcc
import dash_html_components as html

from pyfileconfgui.component import PFCGuiComponent
from pyfileconfgui.dash_ext.component import DashPythonComponent
from pyfileconfgui.dash_ext.python import PythonBlockComponent
from pyfileconfgui.dash_ext.tb import TracebackComponent


class RunEntryComponent(PFCGuiComponent):
    is_running: bool = False
    log_buffer: StringIO = StringIO()

    @property
    def layout(self) -> Sequence[Union['DashPythonComponent', Component]]:
        return [
            html.H2('Running Item'),
            html.P(id='run-input'),
            html.Div(id='run-console-output'),
            dcc.Loading(html.Div(id='run-output')),
            dcc.Interval('run-poll-interval', 500, disabled=True),
        ]

    def add_callbacks(self, app: dash.Dash) -> None:
        from pyfileconfgui.pages.navigator.refresh import REFRESH_INTERVAL_ID
        self.add_callback(
            app,
            self.run_item,
            Output('run-output', 'children'),
            [Input('run-input', 'children')]
        )
        self.add_callback(
            app,
            self.set_polling,
            Output('run-poll-interval', 'disabled'),
            [Input('run-input', 'children'), Input(REFRESH_INTERVAL_ID, 'n_intervals')]
        )
        self.add_callback(
            app,
            self.stream_output,
            Output('run-console-output', 'children'),
            [Input('run-poll-interval', 'n_intervals'), Input(REFRESH_INTERVAL_ID, 'n_intervals')]
        )
        super().add_callbacks(app)

    @property
    def log_output(self) -> str:
        return self.log_buffer.getvalue()

    def run_item(self, path: str):
        self.reset_log()
        self.is_running = True
        orig_stdout = sys.stdout
        sys.stdout = self.log_buffer
        try:
            output = self.gui.runner.run(path)
        except Exception as e:
            sys.stdout = orig_stdout
            self.is_running = False
            return TracebackComponent('run-output-tb').component
        sys.stdout = orig_stdout
        self.is_running = False
        return str(output)

    def set_polling(self, path: str, n_check_intervals: int):
        # Sets disabled for poll interval
        if self.is_running:
            # is running, poll should not be disabled
            return False
        # not running, poll should be disabled
        return True

    def stream_output(self, n_poll_intervals: int, n_check_intervals: int):
        return PythonBlockComponent('run-output-python-block', self.log_output).component

    def reset_log(self):
        self.log_buffer.truncate(0)
        self.log_buffer.seek(0)



from typing import TYPE_CHECKING, Sequence, Union, Mapping, List

import dash
from dash.development.base_component import Component
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output

from pyfileconfgui.dash_ext.component import DashPythonComponent


class RouterComponent(DashPythonComponent):

    def __init__(self, id: str, routes: Mapping[str, DashPythonComponent]):
        self.routes = routes
        super().__init__(id)

    @property
    def layout(self) -> List[Union['DashPythonComponent', Component]]:
        return [
            dcc.Location(id='url', refresh=False),
            html.Div(id='page-content')
        ]

    @property
    def validation_component(self) -> Component:
        route_components = [comp for comp in self.routes.values()]
        component = html.Div(self.layout + route_components)
        return component

    def add_callbacks(self, app: dash.Dash) -> None:
        self.add_callback(
            app,
            self.display_page,
            Output('page-content', 'children'),
            [Input('url', 'pathname')]
        )

    def display_page(self, pathname):
        try:
            comp = self.routes[pathname]
            return comp.component
        except KeyError:
            return '404'



from copy import deepcopy
from typing import Union, List

from pyfileconf import Selector
from pyfileconf.basemodels.collection import Collection
from pyfileconf.logic.get import _get_public_name_or_special_name
from pyfileconf.pipelines.models.collection import PipelineCollection
from pyfileconf.views.object import ObjectView


def full_dict_from_selector(s: Selector) -> dict:
    pc_dict = s._structure.copy()
    out_dict: dict = {}
    pc: Collection
    for manager_key, manager_dict in pc_dict.items():
        out_dict[manager_key] = {}
        for key, pc in manager_dict.items():
            if key == '_general':
                # Skip level of _general dict
                out_dict[manager_key].update(_collection_to_dict(manager_dict['_general']))
            else:
                # Specific class dict, keep level
                out_dict[manager_key][key] = _collection_to_dict(pc)
    return out_dict


def _collection_to_dict(pc: Collection) -> Union[dict, str]:
    out_dict = pc.name_dict.copy()
    value: Union[Collection, ObjectView, List[ObjectView]]
    if len(out_dict) == 1 and all(isinstance(val, ObjectView) for val in out_dict.values()):
        return list(out_dict.values())[0].name
    if (
            len(out_dict) == 1 and
            hasattr(pc, 'klass') and
            pc.klass is not None and
            all(isinstance(val, pc.klass) for val in out_dict.values())
    ):
        return getattr(list(out_dict.values())[0], pc.key_attr)
    for key, value in out_dict.items():
        if isinstance(value, ObjectView):
            # replace with str of name
            out_dict[key] = value.name
        elif isinstance(value, list):
            out_dict[key] = [val.name for val in value]
        elif isinstance(value, Collection):
            # recursively convert PipelineCollection to dict
            out_dict[key] = _collection_to_dict(value)
        elif hasattr(pc, 'klass') and pc.klass is not None and isinstance(value, pc.klass):
            name = getattr(value, pc.key_attr)
            out_dict[key] = name
        else:
            raise ValueError(f'expected one of PipelineCollection, ObjectView, List[ObjectView], or specific class, '
                             f'got {value} of type {type(value)}')
    return out_dict



import os
import shutil
from typing import Callable
from copy import deepcopy

from pyfileconf import PipelineManager

from tests.config import BASE_GENERATED_DIR
from tests.input_files.amodule import a_function
from tests.input_files.mypackage.cmodule import ExampleClass

PM_FOLDER = os.path.join(BASE_GENERATED_DIR, 'pm')
LOGS_PATH = os.path.join(PM_FOLDER, 'MyLogs')
DEFAULTS_FOLDER_NAME = 'defaults'
EXAMPLE_CLASS_FILE_NAME = 'example_class_dict.py'
EXAMPLE_CLASS_FILE_PATH = os.path.join(PM_FOLDER, EXAMPLE_CLASS_FILE_NAME)
EC_CLASS_DICT = {
    'class': ExampleClass,
    'name': 'example_class'
}
CLASS_CONFIG_DICT_LIST = [
    EC_CLASS_DICT
]

PM_DEFAULTS = dict(
    folder=PM_FOLDER,
    name='guipm',
    log_folder=LOGS_PATH,
    default_config_folder_name=DEFAULTS_FOLDER_NAME,
    specific_class_config_dicts=CLASS_CONFIG_DICT_LIST,
)


def full_pm_setup(**kwargs) -> PipelineManager:
    if not os.path.exists(PM_FOLDER):
        os.makedirs(PM_FOLDER)
    write_to_pipeline_dict_file()
    write_example_class_dict_to_file()
    pm = create_pm(**kwargs)
    pm.load()
    return pm


def create_pm(**kwargs) -> PipelineManager:
    all_kwargs = deepcopy(PM_DEFAULTS)
    all_kwargs.update(**kwargs)
    pipeline_manager = PipelineManager(**all_kwargs)
    return pipeline_manager


def delete_pm_project(path: str = PM_FOLDER, logs_path: str = LOGS_PATH):
    all_paths = [
        os.path.join(path, 'defaults'),
        os.path.join(path, 'pipeline_dict.py'),
        logs_path,
    ]
    for path in all_paths:
        if os.path.isdir(path):
            shutil.rmtree(path)
        elif os.path.isfile(path):
            os.remove(path)
        else:
            # Must not exist, maybe add handling for this later
            pass


def write_to_pipeline_dict_file(pm_folder: str = PM_FOLDER, nest_section: bool = True):
    file_path = os.path.join(pm_folder, 'pipeline_dict.py')

    write_str = """
from tests.input_files.amodule import a_function
from tests.input_files.mypackage.cmodule import ExampleClass

pipeline_dict = {
    "my_section": {"stuff": [a_function,], "woo": [ExampleClass,],},
}
"""
    with open(file_path, 'w') as f:
        f.write(write_str)


def class_dict_str(name: str, key: str, value: str) -> str:
    return f'\n{name} = {{\n\t"{key}": ["{value}"],\n}}\n'


def nested_class_dict_str(name: str, section_key: str, key: str, value: str) -> str:
    return f'\n{name} = {{\n\t"{section_key}": {{\n\t\t"{key}": ["{value}"],\n\t}},\n}}\n'


def write_example_class_dict_to_file(nest_section: bool = True):
    if nest_section:
        write_str = nested_class_dict_str('class_dict', 'my_section', 'stuff', 'data')
    else:
        write_str = class_dict_str('class_dict', 'stuff', 'data')
    with open(EXAMPLE_CLASS_FILE_PATH, 'w') as f:
        f.write(write_str)


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--delete', action='store_true')

    args = parser.parse_args()

    if args.delete:
        delete_pm_project()
    else:
        pm = full_pm_setup()



from typing import Union, Sequence

import dash_core_components as dcc
import dash_html_components as html
import dash
from dash.dependencies import Output, Input
from dash.development.base_component import Component

from pyfileconfgui.component import PFCGuiComponent
from pyfileconfgui.dash_ext.component import DashPythonComponent


def edit_item(path: str):
    # TODO [#1]: read file and instantiate editor
    return path


class EditItemComponent(PFCGuiComponent):

    @property
    def layout(self) -> Sequence[Union[DashPythonComponent, Component]]:
        return [
            html.H2('Edit Item'),
            html.P(id='edit-item-name-output'),
            html.Div(id='editor-output'),
        ]

    def add_callbacks(self, app: dash.Dash) -> None:
        self.add_callback(
            app,
            edit_item,
            Output('editor-output', 'children'),
            [Input('edit-item-name-output', 'children')]
        )
        super().add_callbacks(app)



import importlib
import traceback
from typing import Union, Sequence, Optional

import dash
from dash.dependencies import Output, Input, State
from dash.development.base_component import Component
import dash_core_components as dcc
import dash_html_components as html
from pyfileconf.imports.models.statements.obj import ObjectImportStatement

from pyfileconfgui.component import PFCGuiComponent
from pyfileconfgui.dash_ext.component import DashPythonComponent
from pyfileconfgui.dash_ext.tb import TracebackComponent


class CreateEntryComponent(PFCGuiComponent):

    @property
    def layout(self) -> Sequence[Union['DashPythonComponent', Component]]:
        layout = [
            html.H3('Create Item'),
            html.Label('Section Path'),
            dcc.Input(id='section-path-input', placeholder='my.section.path', value=''),
            html.Label('Function/Class Import  (optional)'),
            dcc.Input(id='function-class-import-input', placeholder='from mymod import Stuff', value=''),
            html.Button('Submit', id='create-item-submit-button'),
            html.Div(id='create-item-output')
        ]

        return layout

    def add_callbacks(self, app: dash.Dash) -> None:
        self.add_callback(
            app,
            self.create_item,
            Output('create-item-output', 'children'),
            [Input('create-item-submit-button', 'n_clicks')],
            (State('section-path-input', 'value'), State('function-class-import-input', 'value'))
        )
        super().add_callbacks(app)

    def create_item(self, n_clicks: int, section_path: str, function_class_import: Optional[str]):
        try:
            return self._create_item(n_clicks, section_path, function_class_import)
        except Exception as e:
            return TracebackComponent('create-item-tb').component

    def _create_item(self, n_clicks: int, section_path: str, function_class_import: Optional[str]):
        if not section_path:
            return dash.no_update
        if function_class_import:
            imp = ObjectImportStatement.from_str(function_class_import)
            if len(imp.objs) != 1:
                raise ValueError(f'must have exactly one object import, got {imp.objs}')
            mod = importlib.import_module(imp.module)
            func_or_class = getattr(mod, imp.objs[0])
        else:
            func_or_class = None
        output = self.gui.runner.create(section_path, func_or_class)
        return f'Created {section_path}'



from typing import List


def nested_dict_to_paths(d: dict, begin_path: str = '', sep: str = '/') -> List[str]:
    paths = []
    base_path_parts = [begin_path] if begin_path else []
    for key, value in d.items():
        path_parts = base_path_parts + [key]
        if isinstance(value, str):
            path = sep.join(path_parts + [value])
            paths.append(path)
        elif isinstance(value, list):
            for v in value:
                if not isinstance(v, str):
                    raise ValueError(f'expected only strings inside lists, got {v} of type {type(v)}')
                path = sep.join(path_parts + [v])
                paths.append(path)
        elif isinstance(value, dict):
            nested_begin_path = sep.join(path_parts)
            paths.extend(nested_dict_to_paths(value, begin_path=nested_begin_path, sep=sep))
        else:
            raise ValueError(f'expected only dicts, lists, and strings, got {value} of type {type(value)}')
    return paths


def convert_name_to_id(name: str) -> str:
    return name.replace(' ', '-').replace('_', '-')



from typing import Dict, List

from pyfileconf import Selector

from pyfileconfgui.app import create_app
from pyfileconfgui.component import PFCGuiComponent
from pyfileconfgui.index import add_layout
from pyfileconfgui.pfc.extract import full_dict_from_selector
from pyfileconfgui.pfc.reformat import nested_dict_to_paths
from pyfileconfgui.runner import PFCRunner


class PyFileConfGUI:
    structure: dict
    paths: List[str]
    s: Selector

    def __init__(self):
        self.refresh()
        self.runner = PFCRunner()
        self.app = create_app()
        PFCGuiComponent.register_app(self)
        add_layout(self)

    def run_server(self, **kwargs):
        self.app.run_server(**kwargs)

    @property
    def file_objs(self) -> List[Dict[str, str]]:
        return [{'key': path} for path in self.paths]

    def refresh(self):
        self.s = Selector()
        self.structure = full_dict_from_selector(self.s)
        self.paths = nested_dict_to_paths(self.structure)



from pyfileconfgui.dash_ext.query import get_triggering_id

REFRESH_INTERVAL_ID = 'navigator-refresh-interval'


def is_refresh_trigger():
    trigger_id = get_triggering_id()
    if trigger_id is None or trigger_id == REFRESH_INTERVAL_ID:
        return True
    return False



from typing import Optional

import dash


def get_triggering_id() -> Optional[str]:
    ctx = dash.callback_context

    if not ctx.triggered:
        trigger_id = None
    else:
        trigger_id = ctx.triggered[0]['prop_id'].split('.')[0]

    return trigger_id



from typing import Type, Sequence, Union, Optional, List, Callable

import dash
from dash.dependencies import Output, Input, State
from dash.development.base_component import Component
import dash_html_components as html


class DashPythonComponent:
    app: Optional[dash.Dash] = None
    _callback_output_ids: List[str] = []

    def __init__(self, id: str, **kwargs):
        self.id = id
        self.component_kwargs = kwargs

        if self.app is not None:
            self.add_callbacks(self.app)

    @property
    def component(self) -> html.Div:
        layout = self.layout
        dash_layout = []
        for comp in layout:
            if isinstance(comp, DashPythonComponent):
                dash_comp = comp.component
            else:
                dash_comp = comp
            dash_layout.append(dash_comp)
        return html.Div(dash_layout, id=self.id, **self.component_kwargs)

    @property
    def layout(self) -> Sequence[Union['DashPythonComponent', Component]]:
        raise NotImplementedError

    @classmethod
    def register_app(cls, app: dash.Dash):
        cls.app = app

    def add_callbacks(self, app: dash.Dash) -> None:
        layout = self.layout
        for comp in layout:
            if isinstance(comp, DashPythonComponent):
                comp.add_callbacks(app)

    def add_callback(self, app: dash.Dash, func: Callable, output: Output, inputs: Sequence[Input],
                     state: Sequence[State] = tuple(), prevent_initial_call: Optional[bool] = None) -> None:
        if output.component_id in self._callback_output_ids:
            return
        self._callback_output_ids.append(output.component_id)
        app.callback(output, inputs, state, prevent_initial_call=prevent_initial_call)(func)



from typing import TYPE_CHECKING, Dict

if TYPE_CHECKING:
    from pyfileconfgui.main import PyFileConfGUI

from pyfileconfgui.dash_ext.component import DashPythonComponent


class PFCGuiComponent(DashPythonComponent):
    gui: 'PyFileConfGUI'

    @classmethod
    def register_app(cls, gui: 'PyFileConfGUI'):
        cls.gui = gui
        super().register_app(gui.app)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyfileconfgui.main import PyFileConfGUI

from pyfileconfgui.component import PFCGuiComponent
from pyfileconfgui.dash_ext.router import RouterComponent
from pyfileconfgui.pages.navigator.main import NavigatorComponent


class PFCGUIRouterComponent(PFCGuiComponent, RouterComponent):
    pass


def add_layout(gui: 'PyFileConfGUI'):
    app = gui.app

    nav_component = NavigatorComponent('navigator-root')

    routes = {
        '/': nav_component,
        '/navigator': nav_component
    }

    layout_component = PFCGUIRouterComponent('layout-root', routes)

    app.layout = layout_component.component
    app.validation_layout = layout_component.validation_component

